==> mandrel-23.0.0.0-Final.tar.gz: mandrel-mandrel-23.0.0.0-Final/wasm/src/org.graalvm.wasm/src/org/graalvm/wasm/predefined/BuiltinModule.java <==
/*
 * Copyright (c) 2019, 2022, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * The Universal Permissive License (UPL), Version 1.0
 *
 * Subject to the condition set forth below, permission is hereby granted to any
 * person obtaining a copy of this software, associated documentation and/or
 * data (collectively the "Software"), free of charge and under any and all
 * copyright rights in the Software, and any and all patent rights owned or
 * freely licensable by each licensor hereunder covering either (i) the
 * unmodified Software as contributed to or provided by such licensor, or (ii)
 * the Larger Works (as defined below), to deal in both
 *
 * (a) the Software, and
 *
 * (b) any piece of software and/or hardware listed in the lrgrwrks.txt file if
 * one is included with the Software each a "Larger Work" to which the Software
 * is contributed by such licensors),
 *
 * without restriction, including without limitation the rights to copy, create
 * derivative works of, display, perform, and distribute the Software and make,
 * use, sell, offer for sale, import, export, have made, and have sold the
 * Software and the Larger Work(s), and to sublicense the foregoing rights on
 * either these or other terms.
 *
 * This license is subject to the following condition:
 *
 * The above copyright notice and either this complete permission notice or at a
 * minimum a reference to the UPL must be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package org.graalvm.wasm.predefined;

import com.oracle.truffle.api.CompilerAsserts;
import com.oracle.truffle.api.nodes.RootNode;
import org.graalvm.wasm.WasmContext;
import org.graalvm.wasm.WasmFunction;
import org.graalvm.wasm.WasmFunctionInstance;
import org.graalvm.wasm.WasmInstance;
import org.graalvm.wasm.WasmLanguage;
import org.graalvm.wasm.WasmTable;
import org.graalvm.wasm.WasmType;
import org.graalvm.wasm.exception.Failure;
import org.graalvm.wasm.exception.WasmException;
import org.graalvm.wasm.globals.WasmGlobal;
import org.graalvm.wasm.memory.WasmMemory;
import org.graalvm.wasm.predefined.emscripten.EmscriptenModule;
import org.graalvm.wasm.predefined.go.GoModule;
import org.graalvm.wasm.predefined.spectest.SpectestModule;
import org.graalvm.wasm.predefined.testutil.TestutilModule;
import org.graalvm.wasm.predefined.wasi.WasiModule;

import java.util.HashMap;
import java.util.Map;

public abstract class BuiltinModule {
    private static final Map<String, BuiltinModule> predefinedModules = new HashMap<>();

    static {
        final Map<String, BuiltinModule> pm = predefinedModules;
        pm.put("emscripten", new EmscriptenModule());
        pm.put("testutil", new TestutilModule());
        pm.put("wasi_snapshot_preview1", new WasiModule());
        pm.put("spectest", new SpectestModule());
        pm.put("go", new GoModule());
    }

    public static WasmInstance createBuiltinInstance(WasmLanguage language, WasmContext context, String name, String predefinedModuleName) {
        CompilerAsserts.neverPartOfCompilation();
        final BuiltinModule builtinModule = predefinedModules.get(predefinedModuleName);
        if (builtinModule == null) {
            throw WasmException.create(Failure.UNSPECIFIED_INVALID, "Unknown predefined module: " + predefinedModuleName);
        }
        return builtinModule.createInstance(language, context, name);
    }

    protected abstract WasmInstance createInstance(WasmLanguage language, WasmContext context, String name);

    protected void defineExportedFunction(WasmInstance instance, String name, byte[] paramType, byte[] retTypes, WasmFunctionInstance functionInstance) {
        final int typeIdx = instance.symbolTable().allocateFunctionType(paramType, retTypes, instance.context().getContextOptions().supportMultiValue());
        final WasmFunction function = instance.symbolTable().declareExportedFunction(typeIdx, name);
        instance.setFunctionInstance(function.index(), functionInstance);
        instance.setTarget(function.index(), functionInstance.target());
    }

    protected void defineFunction(WasmInstance instance, String name, byte[] paramTypes, byte[] retTypes, RootNode rootNode) {
        // We could check if the same function type had already been allocated,
        // but this is just an optimization, and probably not very important,
        // since predefined modules have a relatively small size.
        final int typeIdx = instance.symbolTable().allocateFunctionType(paramTypes, retTypes, instance.context().getContextOptions().supportMultiValue());
        final WasmFunction function = instance.symbolTable().declareExportedFunction(typeIdx, name);
        instance.setTarget(function.index(), rootNode.getCallTarget());
    }

    protected int defineExternalGlobal(WasmInstance instance, String globalName, WasmGlobal global) {
        int index = instance.symbolTable().numGlobals();
        instance.symbolTable().declareExportedExternalGlobal(globalName, index, global);
        return index;
    }

    protected int defineGlobal(WasmInstance instance, String name, byte valueType, byte mutability, long value) {
        int index = instance.symbolTable().numGlobals();
        instance.symbolTable().declareExportedGlobalWithValue(name, index, valueType, mutability, value);
        return index;
    }

    protected void defineExternalTable(WasmInstance instance, String tableName, WasmTable externalTable) {
        final boolean referenceTypes = instance.context().getContextOptions().supportBulkMemoryAndRefTypes();
        int index = instance.symbolTable().tableCount();
        instance.symbolTable().allocateExternalTable(index, externalTable, referenceTypes);
        instance.symbolTable().exportTable(index, tableName);
    }

    protected int defineTable(WasmInstance instance, String tableName, int initSize, int maxSize, byte type) {
        final boolean referenceTypes = instance.context().getContextOptions().supportBulkMemoryAndRefTypes();
        switch (type) {
            case WasmType.FUNCREF_TYPE:
                break;
            case WasmType.EXTERNREF_TYPE:
                if (!referenceTypes) {
                    throw WasmException.create(Failure.UNSPECIFIED_MALFORMED, "Only function types are currently supported in tables.");
                }
                break;
            default:
                throw WasmException.create(Failure.MALFORMED_REFERENCE_TYPE, "Only reference types supported in tables.");
        }
        int index = instance.symbolTable().tableCount();
        instance.symbolTable().allocateTable(index, initSize, maxSize, type, referenceTypes);
        instance.symbolTable().exportTable(index, tableName);
        return index;
    }

    protected void defineExternalMemory(WasmInstance instance, String memoryName, WasmMemory externalMemory) {
        instance.symbolTable().allocateExternalMemory(externalMemory);
        instance.symbolTable().exportMemory(memoryName);
    }

    protected void defineMemory(WasmInstance instance, String memoryName, int initSize, int maxSize, boolean is64Bit) {
        instance.symbolTable().allocateMemory(initSize, maxSize, is64Bit);
        instance.symbolTable().exportMemory(memoryName);
    }

    protected void importFunction(WasmInstance instance, String importModuleName, String importFunctionName, byte[] paramTypes, byte[] retTypes, String exportName) {
        final int typeIdx = instance.symbolTable().allocateFunctionType(paramTypes, retTypes, instance.context().getContextOptions().supportMultiValue());
        final WasmFunction function = instance.symbolTable().importFunction(importModuleName, importFunctionName, typeIdx);
        instance.symbolTable().exportFunction(function.index(), exportName);
    }

    protected void importMemory(WasmInstance instance, String importModuleName, String memoryName, int initSize, long maxSize, boolean is64Bit) {
        instance.symbolTable().importMemory(importModuleName, memoryName, initSize, maxSize, is64Bit);
    }

    protected byte[] types(byte... args) {
        return args;
    }
}

==> mandrel-23.1.5.0-Final.tar.gz: mandrel-mandrel-23.1.5.0-Final/wasm/src/org.graalvm.wasm/src/org/graalvm/wasm/predefined/BuiltinModule.java <==
/*
 * Copyright (c) 2019, 2023, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * The Universal Permissive License (UPL), Version 1.0
 *
 * Subject to the condition set forth below, permission is hereby granted to any
 * person obtaining a copy of this software, associated documentation and/or
 * data (collectively the "Software"), free of charge and under any and all
 * copyright rights in the Software, and any and all patent rights owned or
 * freely licensable by each licensor hereunder covering either (i) the
 * unmodified Software as contributed to or provided by such licensor, or (ii)
 * the Larger Works (as defined below), to deal in both
 *
 * (a) the Software, and
 *
 * (b) any piece of software and/or hardware listed in the lrgrwrks.txt file if
 * one is included with the Software each a "Larger Work" to which the Software
 * is contributed by such licensors),
 *
 * without restriction, including without limitation the rights to copy, create
 * derivative works of, display, perform, and distribute the Software and make,
 * use, sell, offer for sale, import, export, have made, and have sold the
 * Software and the Larger Work(s), and to sublicense the foregoing rights on
 * either these or other terms.
 *
 * This license is subject to the following condition:
 *
 * The above copyright notice and either this complete permission notice or at a
 * minimum a reference to the UPL must be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package org.graalvm.wasm.predefined;

import com.oracle.truffle.api.CompilerAsserts;
import com.oracle.truffle.api.nodes.RootNode;
import org.graalvm.wasm.WasmContext;
import org.graalvm.wasm.WasmFunction;
import org.graalvm.wasm.WasmFunctionInstance;
import org.graalvm.wasm.WasmInstance;
import org.graalvm.wasm.WasmLanguage;
import org.graalvm.wasm.WasmTable;
import org.graalvm.wasm.WasmType;
import org.graalvm.wasm.exception.Failure;
import org.graalvm.wasm.exception.WasmException;
import org.graalvm.wasm.globals.WasmGlobal;
import org.graalvm.wasm.memory.WasmMemory;
import org.graalvm.wasm.predefined.emscripten.EmscriptenModule;
import org.graalvm.wasm.predefined.go.GoModule;
import org.graalvm.wasm.predefined.spectest.SpectestModule;
import org.graalvm.wasm.predefined.testutil.TestutilModule;
import org.graalvm.wasm.predefined.wasi.WasiModule;

import java.util.HashMap;
import java.util.Map;

public abstract class BuiltinModule {
    private static final Map<String, BuiltinModule> predefinedModules = new HashMap<>();

    static {
        final Map<String, BuiltinModule> pm = predefinedModules;
        pm.put("emscripten", new EmscriptenModule());
        pm.put("testutil", new TestutilModule());
        pm.put("wasi_snapshot_preview1", new WasiModule());
        pm.put("spectest", new SpectestModule());
        pm.put("go", new GoModule());
    }

    public static WasmInstance createBuiltinInstance(WasmLanguage language, WasmContext context, String name, String predefinedModuleName) {
        CompilerAsserts.neverPartOfCompilation();
        final BuiltinModule builtinModule = predefinedModules.get(predefinedModuleName);
        if (builtinModule == null) {
            throw WasmException.create(Failure.UNSPECIFIED_INVALID, "Unknown predefined module: " + predefinedModuleName);
        }
        return builtinModule.createInstance(language, context, name);
    }

    protected abstract WasmInstance createInstance(WasmLanguage language, WasmContext context, String name);

    protected void defineExportedFunction(WasmInstance instance, String name, byte[] paramType, byte[] retTypes, WasmFunctionInstance functionInstance) {
        final int typeIdx = instance.symbolTable().allocateFunctionType(paramType, retTypes, instance.context().getContextOptions().supportMultiValue());
        final WasmFunction function = instance.symbolTable().declareExportedFunction(typeIdx, name);
        instance.setFunctionInstance(function.index(), functionInstance);
        instance.setTarget(function.index(), functionInstance.target());
    }

    protected void defineFunction(WasmInstance instance, String name, byte[] paramTypes, byte[] retTypes, RootNode rootNode) {
        // We could check if the same function type had already been allocated,
        // but this is just an optimization, and probably not very important,
        // since predefined modules have a relatively small size.
        final int typeIdx = instance.symbolTable().allocateFunctionType(paramTypes, retTypes, instance.context().getContextOptions().supportMultiValue());
        final WasmFunction function = instance.symbolTable().declareExportedFunction(typeIdx, name);
        instance.setTarget(function.index(), rootNode.getCallTarget());
    }

    protected int defineExternalGlobal(WasmInstance instance, String globalName, WasmGlobal global) {
        int index = instance.symbolTable().numGlobals();
        instance.symbolTable().declareExportedExternalGlobal(globalName, index, global);
        return index;
    }

    protected int defineGlobal(WasmInstance instance, String name, byte valueType, byte mutability, long value) {
        int index = instance.symbolTable().numGlobals();
        instance.symbolTable().declareExportedGlobalWithValue(name, index, valueType, mutability, value);
        return index;
    }

    protected void defineExternalTable(WasmInstance instance, String tableName, WasmTable externalTable) {
        final boolean referenceTypes = instance.context().getContextOptions().supportBulkMemoryAndRefTypes();
        int index = instance.symbolTable().tableCount();
        instance.symbolTable().allocateExternalTable(index, externalTable, referenceTypes);
        instance.symbolTable().exportTable(index, tableName);
    }

    protected int defineTable(WasmInstance instance, String tableName, int initSize, int maxSize, byte type) {
        final boolean referenceTypes = instance.context().getContextOptions().supportBulkMemoryAndRefTypes();
        switch (type) {
            case WasmType.FUNCREF_TYPE:
                break;
            case WasmType.EXTERNREF_TYPE:
                if (!referenceTypes) {
                    throw WasmException.create(Failure.UNSPECIFIED_MALFORMED, "Only function types are currently supported in tables.");
                }
                break;
            default:
                throw WasmException.create(Failure.MALFORMED_REFERENCE_TYPE, "Only reference types supported in tables.");
        }
        int index = instance.symbolTable().tableCount();
        instance.symbolTable().allocateTable(index, initSize, maxSize, type, referenceTypes);
        instance.symbolTable().exportTable(index, tableName);
        return index;
    }

    protected void defineExternalMemory(WasmInstance instance, String memoryName, WasmMemory externalMemory) {
        final boolean multiMemory = instance.context().getContextOptions().supportMultiMemory();
        int index = instance.symbolTable().memoryCount();
        instance.symbolTable().allocateExternalMemory(index, externalMemory, multiMemory);
        instance.symbolTable().exportMemory(index, memoryName);
    }

    protected void defineMemory(WasmInstance instance, String memoryName, int initSize, int maxSize, boolean is64Bit, boolean isShared) {
        int index = instance.symbolTable().memoryCount();
        // set multiMemory flag to true, since spectest module has multiple memories
        instance.symbolTable().allocateMemory(index, initSize, maxSize, is64Bit, isShared, true);
        instance.symbolTable().exportMemory(index, memoryName);
    }

    protected void importFunction(WasmInstance instance, String importModuleName, String importFunctionName, byte[] paramTypes, byte[] retTypes, String exportName) {
        final int typeIdx = instance.symbolTable().allocateFunctionType(paramTypes, retTypes, instance.context().getContextOptions().supportMultiValue());
        final WasmFunction function = instance.symbolTable().importFunction(importModuleName, importFunctionName, typeIdx);
        instance.symbolTable().exportFunction(function.index(), exportName);
    }

    protected void importMemory(WasmInstance instance, String importModuleName, String memoryName, int initSize, long maxSize, boolean is64Bit, boolean isShared) {
        final boolean multiMemory = instance.context().getContextOptions().supportMultiMemory();
        int index = instance.symbolTable().memoryCount();
        instance.symbolTable().importMemory(importModuleName, memoryName, index, initSize, maxSize, is64Bit, isShared, multiMemory);
    }

    protected byte[] types(byte... args) {
        return args;
    }
}
